"""

cd {{repo_root}}
python -m lumo.data supervised

python -m lumo.data selfsupervised

python -m lumo.data semisupervised
from datasets.objdect

python -m lumo.data https://github.com/example/dataset objdect


from datasets.objdect
"""
